#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the calculation of various chemical properties such as AO, MO, Transition densities or linear response function.
"""

import time
import numpy as np
import scipy as sp
import rgmol
from rgmol.objects import *
from rgmol.threading_functions import *
from rgmol.grid import *

def gaussian_s(r,contraction_coefficients,exponent_primitives,r0):
    """
    gaussian_s(r,contraction_coefficients,exponent_primitives,r0)

    Returns the calculations for the gaussian s atomic orbital

    Parameters
    ----------
        r : 3d ndarray
            the volume on which the function is calculated
        contraction_coefficients : list
            the contraction coefficient
        exponent_primitives : list
            the exponential factor
        r0 : ndarray of length 3
            the center of the orbital

    Returns
    -------
        s_orbital : ndarray same shape as r

    """

    sum_gaussian = np.zeros(np.shape(r)[1:])
    x,y,z = r
    x0,y0,z0 = r0
    x,y,z = x-x0,y-y0,z-z0
    r_r = x**2 + y**2 + z**2

    for coefs in zip(contraction_coefficients,exponent_primitives):
        sum_gaussian +=  coefs[0] * np.exp(-coefs[1]*r_r)


    return sum_gaussian,

def gaussian_p(r,contraction_coefficients,exponent_primitives,r0):
    """
    gaussian_p(r,contraction_coefficients,exponent_primitives,r0)

    Returns the calculations for the gaussian p atomic orbital
    The order of the orbitals follows the molden format

    Parameters
    ----------
        r : 3d ndarray
            the volume on which the function is calculated
        contraction_coefficients : list
            the contraction coefficients
        exponent_primitives : list
            the exponential factors
        r0 : ndarray of length 3
            the center of the orbital

    Returns
    -------
        p_x_orbital,
        p_y_orbital,
        p_z_orbital
    """

    sum_gaussian_x = np.zeros(np.shape(r)[1:])
    sum_gaussian_y = np.zeros(np.shape(r)[1:])
    sum_gaussian_z = np.zeros(np.shape(r)[1:])
    x,y,z = r
    x0,y0,z0 = r0
    x,y,z = x-x0,y-y0,z-z0

    r_r = x**2 + y**2 + z**2

    for coefs in zip(contraction_coefficients,exponent_primitives):
        sum_gaussian_x += x*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_y += y*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_z += z*coefs[0] * np.exp(-coefs[1]*r_r)


    return sum_gaussian_x,sum_gaussian_y,sum_gaussian_z


def gaussian_d(r,contraction_coefficients,exponent_primitives,r0):
    """
    gaussian_d(r,contraction_coefficients,exponent_primitives,r0)

    Returns the calculations for the gaussian d atomic orbital
    The order of the orbitals follows the molden format

    Parameters
    ----------
        r : 3d ndarray
            the volume on which the function is calculated
        contraction_coefficients : list
            the contraction coefficients
        exponent_primitives : list
            the exponential factors
        r0 : ndarray of length 3
            the center of the orbital

    Returns
    -------
        d_zz_orbital,
        d_xz_orbital,
        d_yz_orbital,
        d_xx_yy_orbital,
        d_xy_orbital
    """

    sum_gaussian_xy = np.zeros(np.shape(r)[1:])
    sum_gaussian_xz = np.zeros(np.shape(r)[1:])
    sum_gaussian_yz = np.zeros(np.shape(r)[1:])
    sum_gaussian_xx_yy = np.zeros(np.shape(r)[1:])
    sum_gaussian_zz = np.zeros(np.shape(r)[1:])
    x,y,z = r
    x0,y0,z0 = r0
    x,y,z = x-x0,y-y0,z-z0
    r0_reshaped = np.array(r0).reshape(3,1,1,1)

    x_x = x*x
    y_y = y*y
    r_r = x**2 + y**2 + z**2

    for coefs in zip(contraction_coefficients,exponent_primitives):
        sum_gaussian_xy += (x*y)*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xz += (x*z)*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_yz += (y*z)*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xx_yy += (x_x-y_y)/2*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zz += (z*z-x_x/2-y_y/2)/3**(1/2) * coefs[0] * np.exp(-coefs[1]*r_r)

    return sum_gaussian_zz, sum_gaussian_xz, sum_gaussian_yz, sum_gaussian_xx_yy, sum_gaussian_xy



def gaussian_f(r,contraction_coefficients,exponent_primitives,r0):
    """
    gaussian_f(r,contraction_coefficients,exponent_primitives,r0)

    Returns the calculations for the gaussian f atomic orbital
    The order of the orbitals follows the molden format

    Parameters
    ----------
        r : 3d ndarray
            the volume on which the function is calculated
        contraction_coefficients : list
            the contraction coefficients
        exponent_primitives : list
            the exponential factors
        r0 : ndarray of length 3
            the center of the orbital

    Returns
    -------
        f_zzz_zrr_orbital,
        f_xzz_xrr_orbital,
        f_yzz_yrr_orbital,
        f_zxx_zyy_orbital,
        f_xyz_orbital,
        f_xxx_xyy_orbital,
        f_yyy_xxy_orbital
    """
    sum_gaussian_yyy_xxy = np.zeros(np.shape(r)[1:])
    sum_gaussian_xyz = np.zeros(np.shape(r)[1:])
    sum_gaussian_yzz_yrr = np.zeros(np.shape(r)[1:])
    sum_gaussian_zzz_zrr = np.zeros(np.shape(r)[1:])
    sum_gaussian_xzz_xrr = np.zeros(np.shape(r)[1:])
    sum_gaussian_zxx_zyy = np.zeros(np.shape(r)[1:])
    sum_gaussian_xxx_xyy = np.zeros(np.shape(r)[1:])
    x,y,z = r
    x0,y0,z0 = r0
    x,y,z = x-x0,y-y0,z-z0
    r0_reshaped = np.array(r0).reshape(3,1,1,1)

    x_x = x*x
    y_y = y*y
    z_z = z*z
    r_r = x**2 + y**2 + z**2

    for coefs in zip(contraction_coefficients,exponent_primitives):
        sum_gaussian_yyy_xxy += (y_y-3*x_x)/(2*6**(1/2))*y*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xyz += (x*y*z)*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_yzz_yrr += (5*z_z-r_r)/(2*10**(1/2))*y*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zzz_zrr += (5*z_z-3*r_r) /(2*15**(1/2))  *z*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xzz_xrr += (5*z_z-r_r)/(2*10**(1/2))*x*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zxx_zyy += (x_x-y_y)/2*z*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xxx_xyy += (3*y_y-x_x)/(2*6**(1/2))*x*coefs[0] * np.exp(-coefs[1]*r_r)


    return sum_gaussian_zzz_zrr, sum_gaussian_xzz_xrr, sum_gaussian_yzz_yrr, sum_gaussian_zxx_zyy, sum_gaussian_xyz, sum_gaussian_xxx_xyy, sum_gaussian_yyy_xxy




def gaussian_g(r,contraction_coefficients,exponent_primitives,r0):
    """
    gaussian_g(r,contraction_coefficients,exponent_primitives,r0)

    Returns the calculations for the gaussian g atomic orbital
    The order of the orbitals follows the molden format

    Parameters
    ----------
        r : 3d ndarray
            the volume on which the function is calculated
        contraction_coefficients : list
            the contraction coefficients
        exponent_primitives : list
            the exponential factors
        r0 : ndarray of length 3
            the center of the orbital

    Returns
    -------
        g_zzzz_orbital,
        g_zzzx_orbital,
        g_zzzy_orbital,
        g_zz_xx_yy_orbital,
        g_zzxy_orbital,
        g_zxxx_orbital,
        g_zyyy_orbital,
        g_xxxx_yyyy_orbital,
        g_xy_xx_yy_orbital
    """
    sum_gaussian_zzzz = np.zeros(np.shape(r)[1:])
    sum_gaussian_zzzy = np.zeros(np.shape(r)[1:])
    sum_gaussian_zzzx = np.zeros(np.shape(r)[1:])
    sum_gaussian_zzxy = np.zeros(np.shape(r)[1:])
    sum_gaussian_zz_xx_yy = np.zeros(np.shape(r)[1:])
    sum_gaussian_zyyy = np.zeros(np.shape(r)[1:])
    sum_gaussian_zxxx = np.zeros(np.shape(r)[1:])
    sum_gaussian_xy_xx_yy = np.zeros(np.shape(r)[1:])
    sum_gaussian_xxxx_yyyy = np.zeros(np.shape(r)[1:])
    x,y,z = r
    x0,y0,z0 = r0
    x,y,z = x-x0,y-y0,z-z0
    r0_reshaped = np.array(r0).reshape(3,1,1,1)

    x_x = x*x
    y_y = y*y
    z_z = z*z
    r_r = x**2 + y**2 + z**2

    for coefs in zip(contraction_coefficients,exponent_primitives):
        sum_gaussian_zzzz += (35*z_z*z_z - 30*z_z*r_r + 3*r_r*r_r)/(2*(16*35)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zzzy += y*z*(7*z_z-3*r_r)/(2*(2*7)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zzzx += x*z*(7*z_z-3*r_r)/(2*(2*7)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zzxy += x*y*(7*z_z-r_r)/(2*(7)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zz_xx_yy += (x_x-y_y)*(7*z_z-r_r)/(2*(2*2*7)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zyyy += y*z*(y_y-3*x_x)/(2*(2)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_zxxx += x*z*(3*y_y-x_x)/(2*(2)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xy_xx_yy += x*y*(y_y-x_x)/(2)*coefs[0] * np.exp(-coefs[1]*r_r)
        sum_gaussian_xxxx_yyyy += (6*x_x*y_y-x_x*x_x-y_y*y_y)/(2*(16)**(1/2))*coefs[0] * np.exp(-coefs[1]*r_r)


    return sum_gaussian_zzzz, sum_gaussian_zzzx, sum_gaussian_zzzy, sum_gaussian_zz_xx_yy, sum_gaussian_zzxy, sum_gaussian_zxxx, sum_gaussian_zyyy, sum_gaussian_xxxx_yyyy, sum_gaussian_xy_xx_yy



def calculate_AO(self):
    """
    calculate_AO()

    Calculate all atomic orbitals for a molecule and puts it in molecule.properties["AO_calculated_list"]
    If no grids were associated with the molecule, it will automatically create all atomic grids

    Parameters
    ----------
        None
            All the data needed should already be inside the molecule object

    Returns
    -------
        AO_calculated_list : list of ndarray
    """

    if not self.mol_grids:
        create_grid_from_mol(self)

    mol_grids = self.mol_grids
    N_grids = len(mol_grids.grids)

    if not "AO_list" in self.properties:
        raise TypeError("The Atomic Orbital functions were not found. Please extract them from a molden file")

    AO_list = self.properties["AO_list"]
    AO_type_list = self.properties["AO_type_list"]
    AO_calculated_list = []

    print("###############################")
    print("# Calculating Atomic Orbitals #")
    print("###############################")
    time_before_calc = time.time()

    Ints = []
    for grid in mol_grids.grids:
        r = grid.xyz_coords

        AO_calculated = []

        for AO_atom in zip(self.atoms,AO_list,AO_type_list):
            r0 = AO_atom[0].pos
            for AO in zip(AO_atom[1],AO_atom[2]):
                exponent_primitives,contraction_coefficients=np.array(AO[0]).transpose()
                if AO[1] == "s":
                    AO_calc = gaussian_s(r,contraction_coefficients,exponent_primitives,r0)
                elif AO[1] == "p":
                    AO_calc = gaussian_p(r,contraction_coefficients,exponent_primitives,r0)
                elif AO[1] == "d":
                    AO_calc = gaussian_d(r,contraction_coefficients,exponent_primitives,r0)
                elif AO[1] == "f":
                    AO_calc = gaussian_f(r,contraction_coefficients,exponent_primitives,r0)
                elif AO[1] == "g":
                    AO_calc = gaussian_g(r,contraction_coefficients,exponent_primitives,r0)
                else: raise ValueError("This type of orbital has not been yet implemented")

                for ao in AO_calc:
                    AO_calculated.append(ao)
                    Ints.append(grid.integrate_product(ao,ao))
        AO_calculated_list.append(np.array(AO_calculated))

    Ints = np.array(Ints).reshape((N_grids,len(AO_calculated)))
    Ints = np.sum(Ints,axis=0).reshape((1,len(AO_calculated),1,1))
    #Renormalize the results. They should be around 1-1e-10 before normalization though
    AO_calculated_list = AO_calculated_list / Ints**(1/2)

    self.properties["AO_calculated_list"] = np.array(AO_calculated_list)

    time_taken = time.time() - time_before_calc
    text_finished = "# Finished Calculating Atomic Orbitals #"
    text_time = "# in {:3.3f} s #".format(time_taken)
    numb_di = max(len(text_finished),len(text_time))
    print("#"*numb_di)
    print(text_finished)
    print(text_time)
    print("#"*numb_di)
    return np.array(AO_calculated_list)



def calculate_MO(self):
    """
    calculate_MO()

    Calculate all molecular orbitals for a molecule and puts it in molecule.properties["MO_calculated_list"]

    If no grids were associated with the molecule, it will automatically create all atomic grids
    If the AO were not calculated it will also calculate them
    By default the MO are normalized to the occupancy of the MO
    That means that in restricted calculations, they are normalized to 2

    Parameters
    ----------
        None
            All the data needed should already be inside the molecule object

    Returns
    -------
        MO_calculated_list : list of ndarray
    """

    if not "AO_calculated_list" in self.properties:
        self.calculate_AO()

    AO_calculated_list = self.properties["AO_calculated_list"]

    MO_calculated_list = []
    MO_list = np.array(self.properties["MO_list"])
    MO_occupancy = self.properties["MO_occupancy"][0]
    N_grids,N_AO,N_r,N_ang = np.shape(AO_calculated_list)


    print('##################################')
    print("# Calculating Molecular Orbitals #")
    print('##################################')

    time_before_calc = time.time()
    Ints = []
    for MO in MO_list:
        MO_not_normalized = np.einsum("ijkl,j->ikl",AO_calculated_list,MO)
        MO_calculated_list.append(MO_not_normalized)
        Ints.append(self.mol_grids.integrate_product(MO_not_normalized,MO_not_normalized))

    N_MO = len(MO_list)

    #Convert to array an transpose to the dimensions : grids,MO,radial,angular
    MO_calculated_list = np.array(MO_calculated_list).transpose((1,0,2,3))
    #Same for int for the two first dimensions
    Ints = np.array(Ints).reshape((1,N_MO,1,1))

    #Renormalize the results. They should be around 1-1e-10 before normalization though
    MO_calculated_list = MO_calculated_list / Ints**(1/2) * MO_occupancy**(1/2)

    self.properties["MO_calculated_list"] = np.array(MO_calculated_list)

    time_taken = time.time() - time_before_calc

    print("###########################################")
    print("# Finished Calculating Molecular Orbitals #")
    print("# in {:3.3f} s #".format(time_taken))
    print("###########################################")

    return np.array(MO_calculated_list)




def calculate_MO_chosen(self,MO_chosen):
    """
    calculate_MO_chosen(MO_chosen)

    Calculate a molecular orbitals for a molecule and puts it in molecule.properties["MO_calculated"][MO_chosen]

    If no grids were associated with the molecule, it will automatically create all atomic grids
    If the AO were not calculated it will also calculate them
    By default the MO are normalized to the occupancy of the MO
    That means that in restricted calculations, they are normalized to 2

    Parameters
    ----------
        MO_chosen : int
            the number of the molecular orbital starting at 0

    Returns
    -------
        MO_calculated_list : ndarray

    """
    if not "AO_calculated" in self.properties:
        calculate_AO(self)

    if not "MO_calculated_list" in self.properties:
        self.properties["MO_calculated_list"] = [[] for k in range(len(self.properties["MO_list"]))]

    AO_calculated = self.properties["AO_calculated"]
    N_AO = len(AO_calculated)
    MO_chosen_calculated = self.properties["MO_calculated_list"][MO_chosen]
    if type(MO_chosen_calculated) is not list:
        return MO_chosen_calculated

    voxel_matrix = self.properties["voxel_matrix"]
    dV = voxel_matrix[0][0] * voxel_matrix[1][1] * voxel_matrix[2][2]

    MO = np.array(self.properties["MO_list"][MO_chosen])
    MO_occupancy = self.properties["MO_occupancy"][0]


    AO_contribution_reshaped = np.array(MO).reshape((N_AO,1,1,1))

    MO_chosen_calculated = np.einsum("ijkl,ijkl->jkl",AO_calculated,AO_contribution_reshaped)

    self.properties["MO_calculated_list"][MO_chosen] = MO_chosen_calculated / (np.einsum("ijk,ijk->",MO_chosen_calculated,MO_chosen_calculated)*dV/MO_occupancy)**(1/2)

    return MO_chosen_calculated / (np.sum(MO_chosen_calculated**2*dV/MO_occupancy)**(1/2))



def calculate_electron_density(self):
    """
    calculate_electron_density()

    Calculates the electron density for a molecule and puts it in molecule.properties["electron_density"]

    If no voxel were associated with the molecule, it will automatically create a voxel
    If the MO were not calculated it will also calculate them

    Parameters
    ----------
        None
            All the data should already be inside the molecule object

    Returns
    -------
        electron_density : ndarray
    """

    if not "MO_calculated_list" in self.properties:
        self.calculate_MO()

    MO = np.array(self.properties["MO_calculated_list"])

    MO_occ = np.array(self.properties["MO_occupancy"])
    MO_occ_arr = MO_occ>0
    MO_occ_index = np.argmin(MO_occ)

    MO_occupied = MO[:,:MO_occ_index]

    electron_density = np.einsum("ijkl,ijkl->ikl",MO_occupied,MO_occupied)

    self.properties["electron_density"] = electron_density
    return electron_density


molecule.calculate_AO = calculate_AO
molecule.calculate_MO = calculate_MO
molecule.calculate_MO_chosen = calculate_MO_chosen
molecule.calculate_electron_density = calculate_electron_density

####################
## EXCITED STATES ##
####################


def calculate_transition_density(self):
    """
    calculate_transition_density()

    Calculates all the transition densities for a molecule
    If the MO were not calculated it will also calculate them

    Parameters
    ----------
        None
            All the data should already be inside the molecule object

    Returns
    -------
        transition_density_list

    Notes
    -----
    The transition densities are defined as

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """

    if not "transition_list" in self.properties:
        raise ValueError("No transitions were found, one should use rgmol.extract_excited_states.extract_transition to extract transition")

    transition_list = self.properties["transition_list"]
    transition_factor_list = self.properties["transition_factor_list"]

    if not "MO_calculated_list" in self.properties:
        self.calculate_MO()

    MO_calculated_list = self.properties["MO_calculated_list"]

    nprocs = rgmol.nprocs
    N_grids,N_MO,N_r,N_ang = np.shape(MO_calculated_list)

    print("##################################")
    print("# Calculating Transition Density #")
    print("##################################")

    time_before_calc = time.time()

    #There should be a way to speed up the calculation
    #This can be done by geting an array of the coeffs with 0 on those who are not there and then doing the einsum
    transi_occ_max = 0
    transi_virt_max = 0
    num_transition = len(transition_list)

    for transition,transition_factor in zip(transition_list,transition_factor_list):
        # print(transition)
        transi_occ = np.max(transition[:,0])
        transi_virt = np.max(transition[:,1])

        if transi_occ_max < transi_occ:
            transi_occ_max = transi_occ
        if transi_virt_max < transi_virt:
            transi_virt_max = transi_virt

    transi_occ_max += 1
    transi_virt_max += 1
    transitions = np.zeros((transi_occ_max,transi_virt_max))
    transition_density_coefficients = np.zeros((num_transition,transi_occ_max,transi_virt_max))
    count = 0
    for transition in range(len(transition_list)):
        for factor_index in range(len(transition_list[transition])):
            occ,virt = transition_list[transition][factor_index]
            transitions[occ,virt] = 1
            transition_density_coefficients[transition,occ,virt] = transition_factor_list[transition][factor_index]



    if nprocs > 1 and 0:
    # if nprocs > 1:
        transition_density_list = calculate_transition_density_multithread(self,transitions,transition_density_coefficients,grid_points,nprocs)

    else:
        transition_density_list = np.zeros((N_grids,len(transition_list),N_r,N_ang))
        for occ in range(transi_occ_max):
            for virt in range(transi_virt_max):
                if transitions[occ,virt]:

                    MO_OCC = MO_calculated_list[:,occ]
                    MO_VIRT = MO_calculated_list[:,virt]
                    MO_product = MO_OCC * MO_VIRT

                    transition_coeffs = transition_density_coefficients[:,occ,virt]
                    for transition in range(num_transition):
                        coeff = transition_coeffs[transition]
                        if coeff!=0:
                            transition_density_list[:,transition] = transition_density_list[:,transition] + coeff * MO_product


    self.properties["transition_density_list"] = transition_density_list


    time_taken = time.time()-time_before_calc
    print("###########################################")
    print("# Finished calculating Transition Density #")
    print('# in {:3.3f} s #'.format(time_taken))
    print("###########################################")
    return transition_density_list


molecule.calculate_transition_density = calculate_transition_density




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the functions used to extract data from orca files.
"""


import codecs
import numpy as np
from rgmol.objects import *
from rgmol.general_function import find_bonds


##########################
## Extraction functions ##
##########################


def _extract_mol(file):
    """
    extracts position and names of atoms
    """
    flag_coords = 0
    flag_getting_coords = 0

    for line in codecs.open(file,'r',encoding="utf-8"):
        if "CARTESIAN COORDINATES (A.U.)" in line:
            Pos = []
            Names = []
            flag_coords = 2
        elif flag_coords:
            flag_coords -= 1
            if not(flag_coords):
                flag_getting_coords = 1

        elif flag_getting_coords:
            if len(line)<3:
                flag_getting_coords = 0
            else:
                lsplit = line.split()
                Pos.append([float(lsplit[5]),float(lsplit[6]),float(lsplit[7])])
                Names.append(lsplit[1])
    return Names,Pos



def extract_properties(file,mol=None):
    """
    extract_properties(file,mol=None)

    Extract the Excited states calculations from an ORCA output and adds the properties to a molecule.


    Parameters
    ----------
        file : str
        mol : molecule object, optional
            If defined, the lists will be put inside the molecule properties

    Returns
    -------
        transition_energy : list
        transition : list
        transition_factor : list
            All the lists are sorted by energy
    """

    flag_states = 0
    flag_completing_state = 0
    flag_completed_state = 0

    flag_spec = 0
    flag_spec_in = 0

    flag_excited_states = 0

    D,P,M = [],[],[]

    for line in codecs.open(file, 'r',encoding="utf-8"):

        if "EXCITED STATES" in line:
            #This will ensure we only keep the last excited states calculations if there is a geom opt
            flag_excited_states = 1
            flag_states = 0
            flag_completing_state = 0
            flag_completed_state = 0
            transition_factor_list = []
            transition_list = []
            transition_energy = []
            flag_states = 1

        elif flag_states:
            lsplit = line.split()
            if len(line)>1 and lsplit[0] == "STATE":
                transition_energy.append(float(lsplit[3]))

                transition = []
                transition_factor = []
                flag_completing_state = 1

            elif flag_completing_state:
                if len(line)==1:
                    flag_completing_state = 0
                    flag_completed_state = 1
                    transition_list.append(np.array(transition))
                    #renormalize factors
                    transition_factor = np.array(transition_factor)
                    # transition_factor = transition_factor / (np.sum(transition_factor**2))**(1/2)

                    transition_factor_list.append(transition_factor)
                else:
                    if "b" in lsplit[0]:
                        raise ValueError("Unrestricted calculations not yet implemented")
                    transition.append([ int(lsplit[0][:-1]),int(lsplit[2][:-1])])
                    fact = lsplit[-1]

                    if len(lsplit) == 5:
                        #TDA FALSE calculation
                        transition_factor.append([ np.sign(float(fact))*abs(float(fact))**(1/2) ])
                    else:
                        #TDA TRUE calculation
                        fact = fact.replace(")","")
                        transition_factor.append([ float(fact) ])

            elif len(line)==1 and flag_completed_state:
                flag_states = 0
                flag_completed_state = 0

        elif "Center of mass = " in line:
            lsplit = line.split()
            num_1,num_2,num_3 = lsplit[-3:]
            num_1, num_2, num_3 = num_1.replace("(",""), num_2.replace("(",""), num_3.replace("(","")
            num_1, num_2, num_3 = num_1.replace(")",""), num_2.replace(")",""), num_3.replace(")","")
            num_1, num_2, num_3 = num_1.replace(",",""), num_2.replace(",",""), num_3.replace(",","")
            num_1,num_2,num_3 = float(num_1),float(num_2),float(num_3)

            center_of_mass = [num_1,num_2,num_3]

        elif "SPECTRUM VIA TRANSITION" in line:
            flag_spec = 4

            if "ABSORPTION" in line and "ELECTRIC DIPOLE" in line:
                abs_ed,abs_vd,cd_ed,cd_vd = 1,0,0,0
                state_transition = []
                energy_transition_spectra = []
            elif "ABSORPTION" in line and "VELOCITY DIPOLE" in line:
                abs_ed,abs_vd,cd_ed,cd_vd = 0,1,0,0
            elif "CD" in line and "ELECTRIC DIPOLE" in line:
                abs_ed,abs_vd,cd_ed,cd_vd = 0,0,1,0
            elif "CD" in line and "VELOCITY DIPOLE" in line:
                abs_ed,abs_vd,cd_ed,cd_vd = 0,0,0,1

        elif flag_spec:
            flag_spec -=1
            if flag_spec == 0: flag_spec_in = 1

        elif flag_spec_in:
            lsplit = line.split()
            if len(line)<3: flag_spec_in = 0
            else:
                if abs_ed:
                    state_transition.append(lsplit[0]+" "+lsplit[1]+" "+lsplit[2])
                    energy_transition_spectra.append(float(lsplit[3]))
                    D.append([float(lsplit[-3]),float(lsplit[-2]),float(lsplit[-1])])
                elif abs_vd:
                    P.append([float(lsplit[-3]),float(lsplit[-2]),float(lsplit[-1])])
                elif cd_ed:
                    M.append([float(lsplit[-3]),float(lsplit[-2]),float(lsplit[-1])])

    if not(flag_excited_states): return None

    array_sort = np.argsort(transition_energy)
    transition_energy_sorted = []
    transition_list_sorted = []
    transition_factor_list_sorted = []
    for sorting in array_sort:
        transition_energy_sorted.append(transition_energy[sorting])
        transition_list_sorted.append(transition_list[sorting])
        transition_factor_list_sorted.append(transition_factor_list[sorting])

    if mol:
        mol.properties["transition_energy"] = transition_energy_sorted
        mol.properties["transition_list"] = transition_list_sorted
        mol.properties["transition_factor_list"] = transition_factor_list_sorted

        mol.properties["state_transition"] = state_transition
        mol.properties["energy_transition_spectra"] = energy_transition_spectra
        mol.properties["center_of_mass"] = center_of_mass
        mol.properties["D"] = D
        mol.properties["P"] = P
        mol.properties["M"] = M

    return transition_energy_sorted,transition_list_sorted,transition_factor_list_sorted

def extract(file,do_order_bonds=0):
    """
    extract(file,do_order_bonds=0)

    Extracts and creates a molecule from an orca output.
    If there is TD-DFT calculations, the molecule will also have the transitions and the dipole moments in its properties.

    Parameters
    ----------
        file : str
            the name of the file
        do_order_bonds : bool, optional
            If the algorithm that finds bonds tries to find the order of the bonds.

    Retruns
    -------
        mol : molecule
            The moelcule
    """

    atom_names,atom_position = _extract_mol(file)


    list_atoms = []
    nicknaming = 0
    for name,pos in zip(atom_names,atom_position):
        atom_x = atom(name,pos,nickname=str(nicknaming))
        list_atoms.append(atom_x)
        nicknaming+=1

    mol = molecule(list_atoms,[],file=file)
    extract_properties(file,mol=mol)

    mol.bonds = find_bonds(mol,do_order_bonds=do_order_bonds)

    return mol

def extract_json(file,mol=None):
    """
    """
    import json

    with open(file) as f:
        data = json.load(f)

    TDDFT = data["Molecule"]["TD-DFT"]

    L = []
    transition_list = []
    for k in TDDFT:
        L.append(k["X+Y"])
        Orbwin = k["OrbWin"]
        transition_list.append(np.array([ [x,y] for x in range(Orbwin[0],Orbwin[1]+1) for y in range(Orbwin[2],Orbwin[3]+1)]))

    N1,N2 = np.shape(np.array(L))
    mol.properties["transition_factor_list"] = np.array(L).reshape((N1,N2,1))
    mol.properties["transition_list"] = transition_list







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds the function that allows to automatically find the bonds for a molecule.
WIP
"""


import numpy as np

dict_number_bonds = {
"H":	1,
"Li":	1,
"Be":	2,
"C":	4,
"N":	3,
"O":	2,
"F":	1,
"Na":	1,
"Mg":	2,
"S":	2,
"Cl":	1,
"Br":   1,
"I":    1
}

dict_electronegativity = {
"H":	2.2,
"Li":	.98,
"Be":	1.57,
"B":	2.04,
"C":	2.55,
"N":	3.04,
"O":	3.44,
"F":	3.98,
"Na":	0.93,
"Mg":	1.31,
"S":	2.58,
"Cl":	3.16,
"Br":   2.96,
"I":    2.66
}


def find_atom_near(molecule,pos,factor_threshold=1.4):
    """
    function that returns the atoms that are near the one of interest
    """

    positions = []

    for atom_x in molecule.atoms:
        positions.append(atom_x.pos)

    positions = np.array(positions)
    distance = np.linalg.norm(positions - pos,axis=1)

    #remove the atom itself
    distance = distance + (distance==0.0)*1e9

    min_distance = np.min(distance)
    return distance < (min_distance * factor_threshold)

def find_nearest_atoms(molecule,pos,number_bonds):
    """
    function that returns the atoms that are near the one of interest
    """

    positions = []

    for atom_x in molecule.atoms:
        positions.append(atom_x.pos)

    positions = np.array(positions)
    distance = np.linalg.norm(positions - pos,axis=1)

    #remove the atom itself
    distance = distance + (distance==0.0)*1e9

    dist_argsort = np.argsort(distance)

    nearest = np.zeros((len(molecule.atoms)))
    for index in dist_argsort[:number_bonds]:
        nearest[index] = 1

    return nearest


def are_bonded(bonds,atom_1,atom_2):
    for bond in range(len(bonds)):
        if bonds[bond][0] == atom_1 and bonds[bond][1] == atom_2:
            return True,bond
        if bonds[bond][0] == atom_2 and bonds[bond][1] == atom_1:
            return True,bond
    return False,0


def is_phenyl(mol,atoms_near):
    """
    Detect if is phenyl if H CC
    """
    H = 1
    C = 2
    flag = 1
    for atom_x in range(len(atoms_near)):
        if atoms_near[atom_x]:
            if mol.atoms[atom_x].name == "C":
                C -= 1
            elif mol.atoms[atom_x].name != "C" and flag:
                #Check the other atom, usually H on the phenyl but can be other. But if a fourth atom, not a phenyl
                H -= 1
                flag = 0
            else: return False
    if C==0 and H==0:
        return True
    return False



def find_bonds(mol,do_order_bonds=0):
    """
    Function that tries to find the bonds of a molecule
    This script is still in WIP
    """
    bonds = []

    number_bonds = [10 for k in range(len(mol.atoms))]
    list_atom_near = []
    list_is_phenyl = []



    #Find number of bonds for each atom
    for atom_x in range(len(mol.atoms)):
        if mol.atoms[atom_x].name in dict_number_bonds:
            number_bonds[atom_x] = dict_number_bonds[mol.atoms[atom_x].name]

    #First pass
    for atom_x in range(len(mol.atoms)):
        atoms_near = find_atom_near(mol,mol.atoms[atom_x].pos)
        if np.sum(atoms_near) >= number_bonds[atom_x]:
            atoms_near = find_nearest_atoms(mol,mol.atoms[atom_x].pos,number_bonds[atom_x])

        for atom_near in range(len(atoms_near)):
            if atoms_near[atom_near] and not(are_bonded(bonds,atom_x,atom_near)[0]):
                bonds.append([atom_x,atom_near,1])
                number_bonds[atom_x] = number_bonds[atom_x]-1
                number_bonds[atom_near] = number_bonds[atom_near]-1




    if do_order_bonds:
        #Second pass where the atoms that only have one neighbor are solved
        for atom_1 in range(len(number_bonds)):
            if mol.atoms[atom_1].name in dict_number_bonds:
                if number_bonds[atom_1] > 0:
                    atom_can_bond = []
                    for atom_2 in range(len(number_bonds)):
                        are_bond, bond_index = are_bonded(bonds,atom_1,atom_2)
                        if are_bond and number_bonds[atom_2]>0:
                            atom_can_bond.append([atom_2,bond_index])

                    if len(atom_can_bond) == 1:
                        #Only one atom, has to be bonded
                        bonds[atom_can_bond[0][1]][2] += number_bonds[atom_1]
                        number_bonds[atom_can_bond[0][0]] -= number_bonds[atom_1]
                        number_bonds[atom_1] = 0

        #Third pass where the bonds that are left are divided equally
        #Do this pass twice
        for j in range(2):
            for atom_1 in range(len(number_bonds)):
                if mol.atoms[atom_1].name in dict_number_bonds:
                    if number_bonds[atom_1] > 0:
                        atom_can_bond = []
                        for atom_2 in range(len(number_bonds)):
                            are_bond, bond_index = are_bonded(bonds,atom_1,atom_2)
                            if are_bond and number_bonds[atom_2]>0:
                                # print(number_bonds[atom_2])
                                atom_can_bond.append([atom_2,bond_index])
                        num_atom_can_bond = len(atom_can_bond)
                        if num_atom_can_bond:
                            bond_number = number_bonds[atom_1] / num_atom_can_bond

                            for atom_2 in range(len(atom_can_bond)):
                                if number_bonds[atom_can_bond[atom_2][0]] < bond_number:
                                    num_atom_can_bond -= 1
                                    bonds[atom_can_bond[atom_2][1]][2] += number_bonds[atom_can_bond[atom_2][0]]
                                    number_bonds[atom_1] -= number_bonds[atom_can_bond[atom_2][0]]
                                    number_bonds[atom_can_bond[atom_2][0]] = 0
                                    bond_number = number_bonds[atom_1] / num_atom_can_bond

                            if int(bond_number*2) == bond_number*2:
                                for atom_2 in range(len(atom_can_bond)):
                                    bonds[atom_can_bond[atom_2][1]][2] += bond_number
                                    number_bonds[atom_can_bond[atom_2][0]] -= bond_number
                                number_bonds[atom_1] = 0


    for atom_1 in range(len(number_bonds)):
        if number_bonds[atom_1] < 0:
            print("The atom {} has too much bonds".format(mol.atoms[atom_1].nickname))
    #add 1 into bonds indexes
    for bond in range(len(bonds)):
        bonds[bond][0]+=1
        bonds[bond][1]+=1


    return bonds












#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the calculation of various chemical properties such as AO, MO, Transition densities or linear response function.
"""

import time
import numpy as np
import scipy as sp
import rgmol
from rgmol.objects import *
from rgmol.threading_functions import *
from rgmol.molecular_calculations import *
from rgmol.grid import *



def reconstruct_AO(mol,grid_points=(80,80,80),delta=5):
    """
    reconstruct_AO(mol,grid_points=(80,80,80),detla=5)

    Calculate all atomic orbitals for a molecule on the representation grid
    If no representation grids were associated with the molecule, it will automatically create it

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, AO_calculated
    """

    if "Reconstructed_AO" in mol.properties:
        return mol.properties["Reconstructed_coords"],mol.properties["Reconstructed_AO"]

    # r,c = create_rectilinear_grid_from_molecule(mol,grid_points=grid_points)
    r,c = create_cubic_grid_from_molecule(mol,grid_points=grid_points,delta=delta)


    if not "AO_list" in mol.properties:
        raise TypeError("The Atomic Orbital functions were not found.")

    AO_list = mol.properties["AO_list"]
    AO_type_list = mol.properties["AO_type_list"]

    AO_calculated = []
    print("##################################")
    print("# Reconstructing Atomic Orbitals #")
    print("#       On Rectilinear Grid      #")
    print("##################################")
    time_before_calc = time.time()

    for AO_atom in zip(mol.atoms,AO_list,AO_type_list):
        r0 = AO_atom[0].pos
        for AO in zip(AO_atom[1],AO_atom[2]):
            exponent_primitives,contraction_coefficients=np.array(AO[0]).transpose()
            if AO[1] == "s":
                AO_calc = gaussian_s(r,contraction_coefficients,exponent_primitives,r0)
            elif AO[1] == "p":
                AO_calc = gaussian_p(r,contraction_coefficients,exponent_primitives,r0)
            elif AO[1] == "d":
                AO_calc = gaussian_d(r,contraction_coefficients,exponent_primitives,r0)
            elif AO[1] == "f":
                AO_calc = gaussian_f(r,contraction_coefficients,exponent_primitives,r0)
            elif AO[1] == "g":
                AO_calc = gaussian_g(r,contraction_coefficients,exponent_primitives,r0)
            else: raise ValueError("This type of orbital has not been yet implemented")

            for ao in AO_calc:
                AO_calculated.append(ao)

    time_taken = time.time() - time_before_calc
    print("###########################################")
    print("# Finished Reconstructing Atomic Orbitals #")
    print("#               in {:3.3f} s              #".format(time_taken))
    print("###########################################")

    mol.properties["Reconstructed_coords"] = c
    mol.properties["Reconstructed_AO"] = np.array(AO_calculated)

    return c,np.array(AO_calculated)


def reconstruct_chosen_MO(mol,MO_chosen,grid_points=(80,80,80),delta=5):
    """
    reconstruct_chosen_MO(mol,grid_points=(80,80,80),delta=5)

    Calculate a chosen molecular orbitals for a molecule on the representation grid
    If no representation grids were associated with the molecule, it will automatically create it
    If the AO were not calculated on this grid, it will also calculate them

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, MO_calculated
    """

    if "Reconstructed_MO" in mol.properties:
        return mol.properties["Reconstructed_coords"],mol.properties["Reconstructed_MO"]

    coords,AO_calculated = reconstruct_AO(mol,grid_points=grid_points,delta=delta)

    MO_list = np.array(mol.properties["MO_list"])
    MO = MO_list[MO_chosen]
    MO_not_normalized = np.einsum("ijkl,i->jkl",AO_calculated,MO)

    return coords,MO_not_normalized


def reconstruct_MO(mol,grid_points=(80,80,80),delta=5):
    """
    reconstruct_MO(mol,grid_points=(80,80,80),delta=5)

    Calculate all molecular orbitals for a molecule on the representation grid
    If no representation grids were associated with the molecule, it will automatically create it
    If the AO were not calculated on this grid, it will also calculate them

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, MO_calculated
    """

    if "Reconstructed_MO" in mol.properties:
        return mol.properties["Reconstructed_coords"],mol.properties["Reconstructed_MO"]

    coords,AO_calculated = reconstruct_AO(mol,grid_points=grid_points,delta=delta)

    occ = mol.properties["MO_occupancy"][0]

    print('#####################################')
    print("# Reconstructing Molecular Orbitals #")
    print("#        On Rectilinear Grid        #")
    print('#####################################')

    time_before_calc = time.time()

    MO_list = np.array(mol.properties["MO_list"])
    MO_calculated = np.einsum("ij,jklm->iklm",MO_list,AO_calculated)
    MO_norm = np.einsum("ijkl,ijkl->i", MO_calculated, MO_calculated) / occ
    MO_norm = MO_norm.reshape((len(MO_norm),1,1,1))
    MO_calculated = MO_calculated / MO_norm **(1/2)


    time_taken = time.time() - time_before_calc

    print("##############################################")
    print("# Finished Reconstructing Molecular Orbitals #")
    print("#                in {:3.3f} s                #".format(time_taken))
    print("##############################################")

    mol.properties["Reconstructed_coords"] = coords
    mol.properties["Reconstructed_MO"] = np.array(MO_calculated)

    return coords,np.array(MO_calculated)


def reconstruct_electron_density(mol,grid_points=(80,80,80),delta=5):
    """
    reconstruct_electron_density(mol,grid_points=(80,80,80),delta=5)

    Calculates the electron density for a molecule on the representation grid
    If no representation grids were associated with the molecule, it will automatically create it
    If the MO were not calculated on this grid, it will also calculate them

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, electron_density
    """

    coords,MO_calculated = reconstruct_MO(mol,grid_points=grid_points,delta=delta)


    MO_occ = np.array(mol.properties["MO_occupancy"])
    MO_occ_arr = MO_occ>0
    MO_occ_index = np.argmin(MO_occ)

    MO_occ = MO_occ[MO_occ_arr]

    MO_occupied = MO_calculated[:MO_occ_index]
    MO_occ_occupied = MO_occ[:MO_occ_index]


    electron_density = np.einsum("ijkl,ijkl->jkl",MO_occupied,MO_occupied)

    mol.properties["Reconstructed_electron_density"] = electron_density

    return coords, electron_density



def reconstruct_fukui_function(mol,mol_p=None,mol_m=None,grid_points=(80,80,80),delta=5):
    """
    reconstruct_fukui_function(mol,mol_p=None,mol_m=None,grid_points=(80,80,80),delta=5)

    Calculates the fukui function using finite differences between electron density on the representation grid.
    If mol_p is provided, f+ will be computed.
    If mol_m is provided, f- will be computed.
    If both are provided, f+, f- and f0 will be computed.
    The fukui functions are automatically added to the molecule.properties with the keys : f0 f+ and f-

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, f0, fp, fm

    Notes
    -----

    :math:`f^+ = \\rho(N+1)-\\rho(N)`
    :math:`f^- = \\rho(N)-\\rho(N-1)`
    :math:`f^0 = (f^+ + f^-)/2`
    """

    if not(mol_p) and not(mol_m):
        raise TypeError("Need at least the molecule with N+1 or N-1 number of electrons")

    if not "Reconstructed_electron_density" in mol.properties:
        coords,elec_dens = reconstruct_electron_density(mol,grid_points=grid_points,delta=delta)
    rhoN = mol.properties["Reconstructed_electron_density"]

    if mol_p:
        #calculate f+
        if not "Reconstructed_electron_density" in mol_p.properties:
            reconstruct_electron_density(mol_p,grid_points=grid_points,delta=delta)

        rhoNp1 = mol_p.properties["Reconstructed_electron_density"]

        fp = rhoNp1 - rhoN
    else:
        fp = None

    if mol_m:
        #calculate f-
        if not "Reconstructed_electron_density" in mol_m.properties:
            reconstruct_electron_density(mol_m,grid_points=grid_points,delta=delta)

        rhoNm1 = mol_m.properties["Reconstructed_electron_density"]

        fm = rhoN - rhoNm1

    else:
        fm = None

    if mol_p and mol_m:
        f0 = (fp+fm)/2
    else: f0 = None

    return coords,f0,fp,fm


####################
## EXCITED STATES ##
####################


def reconstruct_transition_density(mol,grid_points=(80,80,80),delta=5):
    """
    reconstruct_transition_density(mol,grid_points=(80,80,80),delta=5)

    Calculates all the transition densities for a molecule on the representation grid

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        coords, transition_density_list

    Notes
    -----
    The transition densities are defined as

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """

    if not "transition_list" in mol.properties:
        raise ValueError("No transitions were found, one should use rgmol.extract_excited_states.extract_transition to extract transition")

    transition_list = mol.properties["transition_list"]
    transition_factor_list = mol.properties["transition_factor_list"]

    coords,MO_calculated = reconstruct_MO(mol,grid_points=grid_points,delta=delta)

    nprocs = rgmol.nprocs
    nx,ny,nz = grid_points


    print("#####################################")
    print("# Reconstructing Transition Density #")
    print("#####################################")

    time_before_calc = time.time()

    #There should be a way to speed up the calculation
    #This can be done by geting an array of the coeffs with 0 on those who are not there and then doing the einsum
    transi_occ_max = 0
    transi_virt_max = 0
    num_transition = len(transition_list)

    for transition,transition_factor in zip(transition_list,transition_factor_list):
        # print(transition)
        transi_occ = np.max(transition[:,0])
        transi_virt = np.max(transition[:,1])

        if transi_occ_max < transi_occ:
            transi_occ_max = transi_occ
        if transi_virt_max < transi_virt:
            transi_virt_max = transi_virt

    transi_occ_max += 1
    transi_virt_max += 1
    transitions = np.zeros((transi_occ_max,transi_virt_max))
    transition_density_coefficients = np.zeros((num_transition,transi_occ_max,transi_virt_max))
    count = 0
    for transition in range(len(transition_list)):
        for factor_index in range(len(transition_list[transition])):
            occ,virt = transition_list[transition][factor_index]
            transitions[occ,virt] = 1
            transition_density_coefficients[transition,occ,virt] = transition_factor_list[transition][factor_index]
    MO_occ = MO_calculated[:transi_occ_max]


    # if nprocs > 1 and 0:
    if nprocs > 1:
        transition_density_list = calculate_transition_density_multithread(mol,transitions,transition_density_coefficients,grid_points,nprocs)
    else:
        # transition_density_list = np.einsum("ijk,jlmn,klmn->ilmn",transition_density_coefficients,MO_occ,MO_calculated,optimize=True)


        transition_density_list = np.zeros((len(transition_list),nx,ny,nz))
        for occ in range(transi_occ_max):
            for virt in range(transi_virt_max):
                if transitions[occ,virt]:

                    MO_OCC = MO_calculated[occ]
                    MO_VIRT = MO_calculated[virt]
                    MO_product = MO_OCC * MO_VIRT

                    transition_coeffs = transition_density_coefficients[:,occ,virt]
                    for transition in range(num_transition):
                        coeff = transition_coeffs[transition]
                        if coeff!=0:
                            transition_density_list[transition] = transition_density_list[transition] + coeff * MO_product


    mol.properties["Reconstructed_transition_density_list"] = transition_density_list


    time_taken = time.time()-time_before_calc
    print("##############################################")
    print("# Finished Reconstructing Transition Density #")
    print('# in {:3.3f} s #'.format(time_taken))
    print("##############################################")
    return coords,transition_density_list


def reconstruct_chosen_transition_density(mol,chosen_transition_density,grid_points=(80,80,80),delta=5):
    """
    reconstruct_chosen_transition_density(mol,chosen_transition_density,grid_points=(80,80,80),delta=5)

    Calculates a chosen transition density for a molecule on the representation grid

    Parameters
    ----------
        mol : molecule
            the molecule
        chosen_transition_density : int
            the index of the chosen transition_density
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers

    Returns
    -------
        transition_density

    Notes
    -----
    The transition densities are defined as :

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """

    if not "transition_list" in mol.properties:
        raise ValueError("No transitions were found, one should use rgmol.extract_excited_states.extract_transition to extract transition")

    coords,MO_calculated = reconstruct_MO(mol,grid_points=grid_points,delta=delta)

    #Initialize transition density list
    if not "Reconstructed_transition_density_list" in mol.properties:
        mol.properties["Reconstructed_transition_density_list"] = [[] for k in range(len(mol.properties["transition_list"]))]

    if type(mol.properties["Reconstructed_transition_density_list"][chosen_transition_density]) is not list:
        return coords, mol.properties["Reconstructed_transition_density_list"][chosen_transition_density]


    transition_list = mol.properties["transition_list"][chosen_transition_density]
    transition_factor_list = mol.properties["transition_factor_list"][chosen_transition_density]


    transition_factor_list = np.array(transition_factor_list)
    transition_factor_list /= np.sum(transition_factor_list**2)**(1/2)

    nx,ny,nz = grid_points
    transition_density = np.zeros((nx,ny,nz))


    for transition_MO in zip(transition_list, transition_factor_list):

        MO_OCC = MO_calculated[transition_MO[0][0]]
        MO_VIRT = MO_calculated[transition_MO[0][1]]

        transition_density += transition_MO[1][0] * MO_OCC * MO_VIRT

    mol.properties["Reconstructed_transition_density_list"][chosen_transition_density] = transition_density
    return coords,transition_density


def reconstruct_eigenvectors(mol,kernel,grid_points=(80,80,80),delta=5,mol_p=None,mol_m=None,fukui_type="0"):
    """
    reconstruct_eigenvectors(mol,kernel,grid_points=(80,80,80),delta=5,mol_p=None,mol_m=None,fukui_type="0")

    Calculates the eigenmodes of a kernel for a molecule on the representation grid

    Parameters
    ----------
        mol : molecule
            the molecule
        grid_points : list of 3, optional
            the number of points on the representation grid
        delta : float, optional
            the length added on all directions to the box containing all atomic centers
        mol_p : molecule, optional
            The molecule with an electron added. Needed for calculating the softness kernel with a fukui_type of "0" or "+"
        mol_m : molecule, optional
            The molecule with an electron removed. Needed for calculating the softness kernel with a fukui_type of "0" or "-"
        fukui_type : molecule, optional
            The type of fukui function used to calculate the softness kernel. The available types are "0", "+" or "-"

    Returns
    -------
        coords, reconstructed_eigenvectors
    """
    nprocs = rgmol.nprocs
    nx,ny,nz = grid_points
    if kernel == "linear_response_function":
        if "Reconstructed_linear_response_eigenvectors" in mol.properties:
            return mol.properties["Reconstructed_coords"],mol.properties["Reconstructed_linear_response_eigenvectors"]
        contribution_eigenvectors = mol.properties["contribution_linear_response_eigenvectors"]

        coords,Reconstructed_transition_density_list = reconstruct_transition_density(mol,grid_points=grid_points,delta=delta)
        N_trans,nx,ny,nz = np.shape(Reconstructed_transition_density_list)

    if kernel == "softness_kernel":
        if "Reconstructed_softness_kernel_eigenvectors" in mol.properties:
            return mol.properties["Reconstructed_coords"],mol.properties["Reconstructed_softness_kernel_eigenvectors"]
        contribution_eigenvectors = mol.properties["contribution_softness_kernel_eigenvectors"]

        coords,Reconstructed_transition_density_list = reconstruct_transition_density(mol,grid_points=grid_points,delta=delta)
        cooords, f0, fp, fm = reconstruct_fukui_function(mol,mol_p=mol_p,mol_m=mol_m,grid_points=grid_points,delta=delta)

        if "0" in fukui_type:
            Reconstructed_transition_density_list = np.append(Reconstructed_transition_density_list,f0.reshape((1,nx,ny,nz)),axis=0)
        elif "+" in fukui_type or "p" in fukui_type:
            Reconstructed_transition_density_list = np.append(Reconstructed_transition_density_list,fp.reshape((1,nx,ny,nz)),axis=0)
        elif "-" in fukui_type or "m" in fukui_type:
            Reconstructed_transition_density_list = np.append(Reconstructed_transition_density_list,fm.reshape((1,nx,ny,nz)),axis=0)

        N_trans,nx,ny,nz = np.shape(Reconstructed_transition_density_list)


    print("###############################")
    print("# Reconstructing Eigenvectors #")
    print("###############################")
    time_before_calc = time.time()

    if nprocs >1:
        reconstructed_eigenvectors = multithreading_reconstruct_eigenvectors(Reconstructed_transition_density_list,contribution_eigenvectors,nprocs)
    else:
        reconstructed_eigenvectors = []
        for eigenvector in contribution_eigenvectors:
            eigenvector_reshaped = eigenvector.reshape((N_trans,1,1,1))
            reconstructed_eigenvector = np.einsum("ijkl,ijkl->jkl",eigenvector_reshaped,Reconstructed_transition_density_list)
            reconstructed_eigenvectors.append(reconstructed_eigenvector)
    if kernel == "linear_response_function":
        mol.properties["Reconstructed_linear_response_eigenvectors"] = np.array(reconstructed_eigenvectors)
    elif kernel == "softness_kernel":
        mol.properties["Reconstructed_softness_kernel_eigenvectors"] = np.array(reconstructed_eigenvectors)


    time_taken = time.time()-time_before_calc
    print("########################################")
    print("# Finished Reconstructing Eigenvectors #")
    print('# in {:3.3f} s #'.format(time_taken))
    print("########################################")

    return coords,np.array(reconstructed_eigenvectors)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the examples presented throughout the tutorial.
"""

import rgmol
import rgmol.examples

if __name__=="__main__":
    ##For each of these examples, just change False to True

    ## Visualization of chloromethane from an ADF output
    if False:
        file = rgmol.examples.adf_CH3Cl
        mol = rgmol.extract_adf.extract(file)
        mol.plot_radius()

    ## Visualization of a property : the dual
    if False:
        file = rgmol.examples.adf_CH3Cl
        mol = rgmol.extract_adf.extract(file)
        mol.plot_property("dual",factor=2)

    ## Visualization of a condensed kernel : condensed linear response
    if False:
        file = rgmol.examples.adf_CH3Cl
        mol = rgmol.extract_adf.extract(file)
        mol.plot_condensed_kernel("condensed linear response",factor=.5)

    ## Visualization of a diagonalized condensed kernel : condensed linear response
    if False:
        file = rgmol.examples.adf_CH3Cl
        mol = rgmol.extract_adf.extract(file)
        mol.plot_diagonalized_condensed_kernel("condensed linear response")

    ## Visualization of a cube file
    if False:
        file = rgmol.examples.cube_H2CO_MO59
        mol = rgmol.extract_cube.extract(file,do_order_bonds=1)
        mol.plot_isodensity()

    ## Visualization of Atomic Orbitals
    if False:
        file = rgmol.examples.molden_H2CO
        mol = rgmol.extract_molden.extract(file,do_order_bonds=1)
        mol.plot_AO()

    ## Visualization of Molecular Orbitals
    if False:
        file = rgmol.examples.molden_H2CO
        mol = rgmol.extract_molden.extract(file,do_order_bonds=1)
        mol.plot_MO()

    ## Visualization of Transition Densities
    if False:
        file = rgmol.examples.molden_H2CO
        mol = rgmol.extract_molden.extract(file,do_order_bonds=1)
        rgmol.extract_orca.extract_properties(rgmol.examples.orca_H2CO,mol=mol)
        mol.plot_transition_density()

    ## Visualization of Diagonalized Non-Condensed Kernel : Linear response function
    if False:
        file = rgmol.examples.molden_H2CO
        mol = rgmol.extract_molden.extract(file,do_order_bonds=1)
        rgmol.extract_orca.extract_properties(rgmol.examples.orca_H2CO,mol=mol)
        mol.plot_diagonalized_kernel(kernel="linear_response_function")

    ## Visualization of Diagonalized Non-Condensed Kernel : Softness kernel
    if False:
        file = rgmol.examples.molden_H2CO
        mol = rgmol.extract_molden.extract(file,do_order_bonds=1)
        rgmol.extract_orca.extract_properties(rgmol.examples.orca_H2CO,mol=mol)

        file_m = rgmol.examples.molden_H2CO_m
        mol_m = rgmol.extract_molden.extract(file_m)

        mol.plot_diagonalized_kernel(kernel="softness_kernel",mol_m=mol_m,fukui_type="f-")

    ## Visualization of Dipole moments
    if False:
        file = rgmol.examples.orca_CHFClBr
        mol = rgmol.extract_orca.extract(file,do_order_bonds=1)
        mol.plot_dipole_moment()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Notes
-----

This script adds functions to do quicker calculations using multiprocessing
"""

import numpy as np
import threading as th
from rgmol.objects import *
import rgmol.molecular_calculations

#################################
## OVERLAP MATRIX CALCULATIONS ##
#################################

def _overlap_matrix(mol,overlap_matrix,Psi,length,start_i,start_j):
    """Calculates the overlap matrix as
    (S)ij = \\int Psi_i(r) Psi_j(r) dr
    on a part of the grid :
    from start_i to start_i + length and start_j to start_j + length
    """
    N_grids = len(mol.mol_grids.grids)
    for i in range(length):
        for j in range(i+1):
            index_i,index_j = i + start_i, j + start_j
            overlap = 0
            for grid,index_grid in zip(mol.mol_grids.grids,range(N_grids)):
                overlap += grid.integrate(Psi[index_grid,index_i]*Psi[index_grid,index_j])
            overlap_matrix[index_i,index_j] = overlap
            overlap_matrix[index_j,index_i] = overlap


def _overlap_matrix_double(mol,overlap_matrix,Psi,list_div_element):
    """Just do the calculation for multiple ones. This is used when the number of processors is N = 2^k with k odd"""
    for (start_i,start_j),length in list_div_element:
        _overlap_matrix(mol,overlap_matrix,Psi,length,start_i,start_j)



def _divide_triangle(N,K,l):
    """
    This functions divides the rectangle triangle to be computed into K almost equal rectangle triangle.
    This is used to properly divide the calculation between the threads.
    """
    if K == 0:
        return [[l,N]]
    #Divides into 2*2 parts or 4 parts
    l1 = l + np.array([0,0]) #topleft
    l2 = l + np.array([0,N//2]) #topright
    l3 = l + np.array([N//2,0]) #botleft, adds one if odd dim
    l4 = l + np.array([N//2,N//2]) #botright

    if K==1:
        #Divides into 2*2 parts
        return [[[l1,N//2],[l3,N//2+N%2]]] + [[[l2,N//2+N%2],[l4,N//2+N%2]]]

    if K>=2:
        #Divides into 4 parts, not equals if the length is odd
        return _divide_triangle(N//2,K-2,l1) + _divide_triangle(N//2+N%2,K-2,l2) + _divide_triangle(N//2+N%2,K-2,l3) + _divide_triangle(N//2+N%2,K-2,l4)


def calculate_overlap_matrix_multithread(mol,Psi,nprocs):
    """
    This function separates the calculation into multiple threads automatically.
    As the overlap matrix is symmetric, only the rectangle triangle needs to be computed.
    The triangle is divided into 2^K smaller triangles which are all computed in separate threads.
    K is equal to the highest power of 2 that is lower that the number maximum of threads
    """

    N = len(Psi[0])
    overlap_matrix = np.zeros((N,N))

    L = 2**np.arange(10)
    K = np.argmin(nprocs>=L)-1 #Get the higher number of proc that is a power of 2 that will be used

    list_div = _divide_triangle(N,K,np.array([0,0]))
    list_th = []

    if K%2==0:
        for (start_i,start_j),length in list_div:
            kk = th.Thread(target=_overlap_matrix,args=(mol,overlap_matrix,Psi,length,start_i,start_j))
            list_th.append(kk)
            kk.start()

    else:
        for list_div_element in list_div:
            thread = th.Thread(target=_overlap_matrix_double,args=(mol,overlap_matrix,Psi,list_div_element))
            list_th.append(thread)
            thread.start()

    for thread in list_th:
        thread.join()

    return overlap_matrix

#####################################
## TRANSITION DENSITY CALCUALTIONS ##
#####################################



def _product_MO_calc(mol,MO,list_add_transition,transitions,nocc,nvirt,index_move,transition_density_coefficients):

    num_transitions = len(transition_density_coefficients)

    for index in range(len(transitions)):
        if transitions[index]:
            occ,virt = np.unravel_index(index+index_move,(nocc,nvirt))

            MO_product = MO[occ] * MO[virt]

            transition_coeffs = transition_density_coefficients[:,occ,virt]
            for transition in range(num_transitions):
                coeff = transition_coeffs[transition]
                if coeff!=0:
                    list_add_transition[transition] += coeff * MO_product

def _divide_bool(arr,nprocs):
    """
    Divides an array composed of 0 and 1 into multiple fragments that have the same numbers of 1
    """

    somm_bool = np.sum(arr)

    num_per_procs = somm_bool // nprocs
    list_num_per_procs = np.array([num_per_procs for k in range(nprocs)])
    num_procs_add = int(somm_bool % nprocs)

    list_X = [0 for k in range(nprocs+1)]
    list_num_per_procs[:num_procs_add] += 1
    num_proc = 0
    count = 0

    for index_bool in range(len(arr)):
        count += arr[index_bool]

        if count == list_num_per_procs[num_proc]:
            list_X[num_proc+1] = index_bool+1
            num_proc += 1
            count = 0
            if num_proc == nprocs:
                break

    list_bools = []
    for k in range(nprocs):
        if k!=nprocs-1:
            list_bools.append(arr[list_X[k]:list_X[k+1]])
        else:
            list_bools.append(arr[list_X[k]:])

    return list_bools,list_X


def calculate_transition_density_multithread(mol,transitions,transition_density_coefficients,grid_points,nprocs):
    """
    This function divides the virtual coefficients into multiple parts
    """
    #This hard cap is because it is actually slower to put too much processors
    if nprocs > 4:
        nprocs = 4

    num_transitions = len(transition_density_coefficients)
    nx,ny,nz = grid_points

    list_add_transitions = np.zeros((nprocs,num_transitions,nx,ny,nz))
    list_th = []
    nocc,nvirt = np.shape(transitions)


    list_transitions,list_X = _divide_bool(transitions.ravel(),nprocs)

    for list_transition,index_move,index_proc in zip(list_transitions,list_X,range(nprocs)):
        thread = th.Thread(target=_product_MO_calc,args=(mol,mol.properties["Reconstructed_MO"],list_add_transitions[index_proc],list_transition,nocc,nvirt,index_move,transition_density_coefficients))
        list_th.append(thread)
        thread.start()

    for thread in list_th:
        thread.join()
    transition_density_list = np.einsum("ijklm->jklm",list_add_transitions)

    return transition_density_list


####################################
## Reconstruction of eigenvectors ##
####################################


def _reconstruct(reconstructed_eigenvectors,transitions,eigenvectors,index):

    for eigenvector,index_eig in zip(eigenvectors,range(len(eigenvectors))):
        reconstructed_eigenvector = np.einsum("i,ijkl->jkl",eigenvector,transitions)
        # reconstructed_eigenvector_norm = mol.mol_grids.(reconstructed_eigenvector,reconstructed_eigenvector)
        reconstructed_eigenvectors[index+index_eig] = reconstructed_eigenvector


def _divide_eigenvectors(eigenvectors,nprocs):

    num_transitions = len(eigenvectors)

    num_per_procs = num_transitions // nprocs
    list_num_per_procs = np.array([num_per_procs for k in range(nprocs)])
    num_procs_add = int(num_transitions % nprocs)
    list_num_per_procs[:num_procs_add] += 1


    list_eigenvectors = []
    list_index = []
    count = 0
    for index in range(nprocs):
        list_eigenvectors.append(eigenvectors[count:count+list_num_per_procs[index]])
        list_index.append(count)
        count += list_num_per_procs[index]

    return list_eigenvectors,list_index


def multithreading_reconstruct_eigenvectors(transitions,eigenvectors,nprocs):
    """REconstructs the eigenvectors"""

    #This hard cap is because it is actually slower to put too much processors
    if nprocs > 4:
        nprocs = 4

    N_trans,nx,ny,nz = np.shape(transitions)

    list_eigenvectors,list_index = _divide_eigenvectors(eigenvectors,nprocs)
    reconstructed_eigenvectors = [np.zeros((nx,ny,nz)) for k in range(N_trans)]

    list_th = []
    for eigenvectors,index in zip(list_eigenvectors,list_index):
        thread = th.Thread(target=_reconstruct,args=(reconstructed_eigenvectors,transitions,eigenvectors,index))
        list_th.append(thread)
        thread.start()

    for thread in list_th:
        thread.join()

    return reconstructed_eigenvectors

##
def _reconstruct_atomic_grids(mol,reconstructed_eigenvectors,reconstructed_eigenvector_norms,transitions,eigenvectors,index):

    N_grids,N_trans,N_r,N_ang = np.shape(transitions)

    for eigenvector,index_eig in zip(eigenvectors,range(len(eigenvectors))):
        reconstructed_eigenvector = np.einsum("j,ijkl->ikl",eigenvector,transitions)
        reconstructed_eigenvector_norm = mol.mol_grids.integrate_product(reconstructed_eigenvector,reconstructed_eigenvector)
        reconstructed_eigenvector_norms[index+index_eig] = reconstructed_eigenvector_norm
        reconstructed_eigenvectors[index+index_eig] = reconstructed_eigenvector/reconstructed_eigenvector_norm**(1/2)



def multithreading_reconstruct_eigenvectors_atomic_grids(mol,transitions,eigenvectors,nprocs):
    """REconstructs the eigenvectors"""

    #This hard cap is because it is actually slower to put too much processors
    if nprocs > 4:
        nprocs = 4

    N_grids,N_trans,N_r,N_ang = np.shape(transitions)

    list_eigenvectors,list_index = _divide_eigenvectors(eigenvectors,nprocs)
    reconstructed_eigenvectors = [np.zeros((N_grids,N_r,N_ang)) for k in range(N_trans)]
    reconstructed_eigenvector_norms = [0 for k in range(N_trans)]

    list_th = []
    for eigenvectors,index in zip(list_eigenvectors,list_index):
        thread = th.Thread(target=_reconstruct_atomic_grids,args=(mol,reconstructed_eigenvectors,reconstructed_eigenvector_norms,transitions,eigenvectors,index))
        list_th.append(thread)
        thread.start()

    for thread in list_th:
        thread.join()

    return np.array(reconstructed_eigenvectors),np.array(reconstructed_eigenvector_norms)








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the plotting of chemical properties using pyvista.
"""

import numpy as np
import pyvista
from rgmol.objects import *
import rgmol.molecular_calculations
from ._functions import *

###############################
## Defines class for sliders ##
###############################

class _slider:
    def __init__(self,func,value,cutoff):
        self.func = func
        self.value = value
        self.cutoff = cutoff

    def __call__(self, slider, value):
        if slider == "cutoff":
            self.cutoff = value
        else:
            self.value = value
        self.func(self.value,self.cutoff)




def plot_NOCV(self,list_fragments,grid_points=(100,100,100),delta=10,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plotsNOCV

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        method : str, optional
            The method of calculation of the kernel, only eigenmodes by default. More information can be found in the notes.
        plotting_method : str, optional
            The method used for plotting. By default isodensity. The other methods are : "multiple isodensities", "volume"
        number_eigenvectors : int, optional
            Only used in the methods "total" and "partial". The number of eigenvectors to be computed
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        cutoff : float, optional
            The initial cutoff of the isodensity plot for the isodensity plotting method. By default .2
        number_isodensities : int, optional
            The number of isodensities to be plotted if the method used is multiple isodensities
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function

    """

    if not "NOCV" in self.properties:
        self.calculate_NOCV(list_fragments,grid_points=grid_points,delta=delta)


    eigenvectors = self.properties["NOCV"]
    eigenvalues = self.properties["NOCV_eigenvalues"]

    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_NOCV(value,cutoff):
        vector_number = int(round(value))
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],eigenvectors[vector_number-1],opacity=opacity,factor=factor,cutoff=cutoff)
        plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")

    slider_function = _slider(create_mesh_NOCV,1,cutoff)
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    slider_function = _slider(create_mesh_NOCV,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(eigenvectors)],value=1,title="Eigenvector", fmt="%1.0f")
    plotter.show(full_screen=False)


molecule.plot_NOCV = plot_NOCV
